from .gremlin import *  # noqa


def iter_degrees(db):
    query = db.query
    for item in db.vertices():
        degree = query(outgoings, count)(item)
        degree += query(incomings, count)(item)
        yield item, degree


def iter_neighbors(db, item):
    query = db.query
    for neighbor in query(outgoings, end)(item):
        yield neighbor
    for neighbor in query(incomings, start)(item):
        yield neighbor


def core_number(graphdb):
    degrees = dict(iter_degrees(graphdb))
    # Sort nodes by degree.
    vertices = sorted(degrees, key=degrees.get)
    bin_boundaries = [0]
    curr_degree = 0
    for i, v in enumerate(vertices):
        if degrees[v] > curr_degree:
            bin_boundaries.extend([i] * (degrees[v]-curr_degree))
            curr_degree = degrees[v]
    positions = {v: pos for pos, v in enumerate(vertices)}
    # The initial guess for the core number of a node is its degree.
    core = degrees
    neighbors = {v: set(iter_neighbors(graphdb, v)) for v in vertices}
    for v in vertices:
        for u in neighbors[v]:
            if u in core and v in core and core[u] > core[v]:
                if v in neighbors[u]:
                    neighbors[u].remove(v)
                    position = positions[u]
                    bin_start = bin_boundaries[core[u]]
                    positions[u] = bin_start
                    positions[vertices[bin_start]] = position
                    vertices[bin_start], vertices[position] = vertices[position], vertices[bin_start]  # noqa
                    bin_boundaries[core[u]] += 1
                    core[u] -= 1
    return core


def k_core(db, k=None):
    core = core_number(db)
    k = k if k else max(core.values()) 
    func = lambda v, k, core: core[v] >= k
    nodes = [v for v in core if func(v, k, core)]
    return nodes
